import subprocess
from pathlib import Path

import pytest

from naas_abi_cli.cli import bootstrap


def test_find_abi_project_root_from_nested_folder(tmp_path: Path) -> None:
    project_root = tmp_path / "project"
    nested_path = project_root / "src" / "pkg"
    nested_path.mkdir(parents=True)
    (project_root / "pyproject.toml").write_text(
        '[project]\nname = "demo"\ndependencies = ["naas-abi-cli"]\n',
        encoding="utf-8",
    )

    resolved_root = bootstrap.find_abi_project_root(nested_path)

    assert resolved_root == project_root


def test_find_abi_project_root_returns_none_for_non_abi_project(tmp_path: Path) -> None:
    project_root = tmp_path / "project"
    nested_path = project_root / "src"
    nested_path.mkdir(parents=True)
    (project_root / "pyproject.toml").write_text(
        '[project]\nname = "demo"\ndependencies = ["requests"]\n',
        encoding="utf-8",
    )

    resolved_root = bootstrap.find_abi_project_root(nested_path)

    assert resolved_root is None


def test_maybe_rerun_in_project_context_skips_when_guard_env_set(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setenv(bootstrap.REEXEC_ENV_VAR, "true")

    assert bootstrap.maybe_rerun_in_project_context(["config", "validate"]) is False


def test_maybe_rerun_in_project_context_executes_uv_run(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_root = tmp_path / "project"
    nested_path = project_root / "src" / "feature"
    nested_path.mkdir(parents=True)
    (project_root / "pyproject.toml").write_text(
        '[project]\nname = "demo"\ndependencies = ["naas-abi-core"]\n',
        encoding="utf-8",
    )

    called: dict[str, object] = {}
    info: dict[str, object] = {}

    def _fake_run(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        called["arguments"] = args[0]
        called["cwd"] = kwargs["cwd"]
        called["check"] = kwargs["check"]
        called["env"] = kwargs["env"]

    monkeypatch.delenv(bootstrap.REEXEC_ENV_VAR, raising=False)
    monkeypatch.delenv("PYTEST_CURRENT_TEST", raising=False)
    monkeypatch.delenv("PYTEST_VERSION", raising=False)
    monkeypatch.setattr(bootstrap.sys, "argv", ["abi"])
    monkeypatch.setattr(bootstrap, "find_abi_project_root", lambda: project_root)
    monkeypatch.setattr(bootstrap.Path, "cwd", lambda: nested_path)
    monkeypatch.setattr(bootstrap, "_get_system_cli_version", lambda: "1.0.0")
    monkeypatch.setattr(bootstrap, "_get_project_cli_version", lambda _: "1.1.0")
    monkeypatch.setattr(
        bootstrap,
        "_show_rerun_info",
        lambda root, system, project: info.update(
            {"root": root, "system": system, "project": project}
        ),
    )
    monkeypatch.setattr(bootstrap.subprocess, "run", _fake_run)

    result = bootstrap.maybe_rerun_in_project_context(["module", "list"])

    assert result is True
    assert called["arguments"] == [
        "uv",
        "run",
        "--project",
        str(project_root),
        "--active",
        "abi",
        "module",
        "list",
    ]
    assert called["cwd"] == str(nested_path)
    assert called["check"] is True
    env = called["env"]
    assert isinstance(env, dict)
    assert env[bootstrap.REEXEC_ENV_VAR] == "true"
    assert info == {"root": project_root, "system": "1.0.0", "project": "1.1.0"}


def test_maybe_rerun_in_project_context_falls_back_when_uv_missing(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_root = tmp_path / "project"
    project_root.mkdir(parents=True)
    (project_root / "pyproject.toml").write_text(
        '[project]\nname = "demo"\ndependencies = ["naas-abi-cli"]\n',
        encoding="utf-8",
    )

    def _missing_uv(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        raise FileNotFoundError("uv not found")

    monkeypatch.delenv(bootstrap.REEXEC_ENV_VAR, raising=False)
    monkeypatch.delenv("PYTEST_CURRENT_TEST", raising=False)
    monkeypatch.delenv("PYTEST_VERSION", raising=False)
    monkeypatch.setattr(bootstrap.sys, "argv", ["abi"])
    monkeypatch.setattr(bootstrap, "find_abi_project_root", lambda: project_root)
    monkeypatch.setattr(bootstrap, "_get_system_cli_version", lambda: "1.0.0")
    monkeypatch.setattr(bootstrap, "_get_project_cli_version", lambda _: "1.1.0")
    monkeypatch.setattr(bootstrap, "_show_rerun_info", lambda *_: None)
    monkeypatch.setattr(bootstrap.subprocess, "run", _missing_uv)

    assert bootstrap.maybe_rerun_in_project_context(["config", "validate"]) is False


def test_maybe_rerun_in_project_context_propagates_reexec_failure(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_root = tmp_path / "project"
    project_root.mkdir(parents=True)
    (project_root / "pyproject.toml").write_text(
        '[project]\nname = "demo"\ndependencies = ["naas-abi-cli"]\n',
        encoding="utf-8",
    )

    def _failing_uv(*args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        raise subprocess.CalledProcessError(returncode=4, cmd=["uv"])

    monkeypatch.delenv(bootstrap.REEXEC_ENV_VAR, raising=False)
    monkeypatch.delenv("PYTEST_CURRENT_TEST", raising=False)
    monkeypatch.delenv("PYTEST_VERSION", raising=False)
    monkeypatch.setattr(bootstrap.sys, "argv", ["abi"])
    monkeypatch.setattr(bootstrap, "find_abi_project_root", lambda: project_root)
    monkeypatch.setattr(bootstrap, "_get_system_cli_version", lambda: "1.0.0")
    monkeypatch.setattr(bootstrap, "_get_project_cli_version", lambda _: "1.1.0")
    monkeypatch.setattr(bootstrap, "_show_rerun_info", lambda *_: None)
    monkeypatch.setattr(bootstrap.subprocess, "run", _failing_uv)

    with pytest.raises(SystemExit, match="4") as error:
        bootstrap.maybe_rerun_in_project_context(["config", "validate"])

    assert error.value.code == 4



import os
import subprocess
import sys
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

import click

REEXEC_ENV_VAR = "LOCAL_UV_RAN"
ABI_PROJECT_MARKERS = (
    "naas-abi-cli",
    "naas-abi-core",
    "naas-abi-marketplace",
    "[tool.uv.sources]",
)


def _is_abi_project(pyproject_path: Path) -> bool:
    try:
        pyproject_content = pyproject_path.read_text(encoding="utf-8").lower()
    except OSError:
        return False

    return any(marker in pyproject_content for marker in ABI_PROJECT_MARKERS)


def find_abi_project_root(start_path: Path | None = None) -> Path | None:
    current_path = (start_path or Path.cwd()).resolve()

    for candidate in (current_path, *current_path.parents):
        pyproject_path = candidate / "pyproject.toml"
        if pyproject_path.exists() and _is_abi_project(pyproject_path):
            return candidate

    return None


def _get_system_cli_version() -> str:
    try:
        return version("naas-abi-cli")
    except PackageNotFoundError:
        return "unknown"


def _get_project_cli_version(project_root: Path) -> str:
    command = [
        "uv",
        "run",
        "--project",
        str(project_root),
        "--active",
        "python",
        "-c",
        "from importlib.metadata import version; print(version('naas-abi-cli'))",
    ]

    try:
        result = subprocess.run(
            command,
            cwd=str(project_root),
            capture_output=True,
            text=True,
            check=True,
            env={**os.environ, REEXEC_ENV_VAR: "true"},
        )
    except (FileNotFoundError, subprocess.CalledProcessError):
        return "unknown"

    project_version = result.stdout.strip()
    return project_version or "unknown"


def _show_rerun_info(
    project_root: Path, system_version: str, project_version: str
) -> None:
    click.secho(
        "[abi] ABI project detected. Re-running in local project environment.",
        fg="yellow",
        bold=True,
    )
    click.secho(
        f"      project: {project_root}",
        fg="bright_yellow",
    )
    click.secho(
        f"      versions: system={system_version} | project={project_version}",
        fg="bright_yellow",
    )


def maybe_rerun_in_project_context(argv: list[str]) -> bool:
    if os.getenv(REEXEC_ENV_VAR):
        return False
    if os.getenv("PYTEST_CURRENT_TEST") is not None or os.getenv("PYTEST_VERSION"):
        return False
    if "pytest" in Path(sys.argv[0]).name:
        return False

    project_root = find_abi_project_root()
    if project_root is None:
        return False

    system_version = _get_system_cli_version()
    project_version = _get_project_cli_version(project_root)
    _show_rerun_info(project_root, system_version, project_version)

    invocation_cwd = Path.cwd()

    arguments = [
        "uv",
        "run",
        "--project",
        str(project_root),
        "--active",
        "abi",
        *argv,
    ]

    try:
        subprocess.run(
            arguments,
            cwd=str(invocation_cwd),
            env={**os.environ, REEXEC_ENV_VAR: "true"},
            check=True,
        )
    except FileNotFoundError:
        return False
    except subprocess.CalledProcessError as error:
        raise SystemExit(error.returncode) from error

    return True



import click
from naas_abi_core.engine.Engine import Engine


@click.group("run")
def run():
    pass


@run.command(
    "script",
    context_settings={
        "ignore_unknown_options": True,
        "allow_extra_args": True,
    },
)
@click.pass_context
@click.argument("path", type=str, required=True)
def run_script(ctx: click.Context, path: str):
    click.echo(f"[abi run script] Loading engine for: {path}")
    engine = Engine()
    engine.load()
    click.echo("[abi run script] Engine loaded. Running script...")

    import runpy
    import sys

    # Forward any arguments after the script path to the target script so
    # argparse/click inside the script can parse them.
    script_args = list(ctx.args)
    sys.argv = [path, *script_args]
    runpy.run_path(path, run_name="__main__")
